import typing


def to_snake_case(value: typing.Any) -> str:
    """Converts the camelCase or PascalCase value into a snake_case value."""
    if value is None or value == "":
        return ""

    cleaned = str(value).replace("-", "_").strip()
    characters = [cleaned[0].lower()]
    for last_index, character in enumerate(cleaned[1:]):
        last_character = cleaned[last_index]
        was_lowercase = last_character.lower() == last_character
        is_lowercase = character.lower() == character
        if last_character == "_":
            characters.append(character.lower())
        elif was_lowercase and not is_lowercase:
            characters.append(f"_{character.lower()}")
        else:
            characters.append(character.lower())
    return "".join(characters).strip("_")



import pathlib as _pathlib
from importlib import metadata as _metadata

import toml as _toml

from aok._operations import to_comparator  # noqa
from aok._comparisons import *  # noqa

try:
    __version__ = _metadata.version(__package__)
except _metadata.PackageNotFoundError:  # pragma: no-cover
    # If the package is not installed such that it has distribution metadata
    # fallback to loading the version from the pyproject.toml file.
    __version__ = _toml.loads(
        _pathlib.Path(__file__).parent.parent.joinpath("pyproject.toml").read_text()
    )["tool"]["poetry"]["version"]



import fnmatch
import re
import typing

import yaml

from aok import _definitions


class Like(_definitions.Comparator):
    """Compares strings using unix-shell wildcard like regexes."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        return fnmatch.fnmatch(observed, self.value)

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Like":
        value: str = loader.construct_python_str(node)
        return cls(value)


class LikeCase(_definitions.Comparator):
    """Compares strings using unix-shell wildcard like regexes."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        return fnmatch.fnmatchcase(observed, self.value)

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "LikeCase":
        value: str = loader.construct_python_str(node)
        return cls(value)


class Match(_definitions.Comparator):
    """Compares strings using the compiled regex."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Determines if the value matches the regular expression."""
        pattern = re.compile(self.value["regex"], flags=self.value.get("flags", 0))
        return pattern.match(observed) is not None

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Match":
        if isinstance(node, yaml.ScalarNode):
            regex = loader.construct_python_str(node)
            value = {"regex": regex}
        else:
            value = loader.construct_mapping(node, deep=True)
        return cls(value)


Like.register()
like = getattr(Like, "constructor", Like)

LikeCase.register()
like_case = getattr(LikeCase, "constructor", LikeCase)

Match.register()
match = getattr(Match, "constructor", Match)

__all__ = [
    "Like",
    "like",
    "LikeCase",
    "like_case",
    "Match",
    "match",
]



import pathlib

import yaml
from pytest import mark

import aok

directory = pathlib.Path(__file__).parent.joinpath("scenarios").absolute()

scenario_paths = [p.name for p in directory.iterdir() if p.name.endswith(".yaml")]


@mark.parametrize("filename", scenario_paths)
def test_scenario(filename: str):
    """Test the expected scenario against the executed aok validation."""
    path = directory.joinpath(filename)
    scenario = yaml.full_load(path.read_text())
    comparator: aok.Okay = scenario["comparator"]
    result = comparator.compare(
        scenario["observed"],
        subset=scenario.get("subset", False),
    )

    expectations = scenario["expected"]
    assert result.success == expectations["success"], result.to_diff_info()
    assert result.failed_keys() == set(expectations.get("failed_keys", []))



import dataclasses
import textwrap
import typing

import yaml
import yaml.constructor

import aok
from aok import _definitions

ArbitraryDict = typing.Dict[typing.Any, typing.Any]


@dataclasses.dataclass()
class Dict(_definitions.Comparator):
    """Main class in which aok assertions are made."""

    def compare(
        self,
        observed: ArbitraryDict,
        subset: bool = False,
    ) -> _definitions.Comparison:
        value = typing.cast(ArbitraryDict, self.value or {})
        other = observed or {}

        keys = set(list(value.keys()))
        if not subset:
            keys.update(set(list(observed.keys())))

        results: typing.Dict[typing.Any, _definitions.Comparison] = {
            key: aok.to_comparator(self.value.get(key)).compare(other.get(key), subset)
            for key in keys
        }

        return _definitions.Comparison(
            operation=self.operation_name(),
            success=all([r.success for r in results.values()]),
            expected=self.value,
            observed=observed,
            children=results,
        )


@dataclasses.dataclass()
class Okay(Dict):
    """Root dictionary object for comparison."""

    def assert_subset(self, observed: ArbitraryDict, message: str = None):
        result = self.compare(observed, subset=True)
        heading = message or "One or more subset differences were found"
        assert result.success, "{}\n{}".format(
            heading,
            textwrap.indent(yaml.safe_dump(result.to_diff_info()), "  "),
        )

    def assert_all(self, observed: ArbitraryDict, message: str = None):
        result = self.compare(observed, subset=False)
        heading = message or "One or more exact differences were found"
        assert result.success, "{}\n{}".format(
            heading,
            textwrap.indent(yaml.safe_dump(result.to_diff_info()), "  "),
        )

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Okay":
        """Loads the dict from a yaml parser."""
        try:
            return cls(value=loader.construct_mapping(node, deep=True))
        except yaml.constructor.ConstructorError:
            if node.value == "":
                return cls(value={})
            raise

    @classmethod
    def register(cls):
        """Overrides the registration in this case for base registration."""
        yaml.add_constructor("!aok", Okay.parse_yaml)


Okay.register()

__all__ = ["Okay", "Dict"]



import typing

from aok import _definitions


class Optional(_definitions.Comparator):
    """Compares two value as an equality or allow None."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Make an optional equality comparison."""
        return self.value is None or self.value == observed


class NotNull(_definitions.Comparator):
    """Won't allow a value to be null."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Make an optional equality comparison."""
        return self.value is not None


NotNull.register()
not_null = getattr(NotNull, "constructor", NotNull)

Optional.register()
optional = getattr(Optional, "constructor", Optional)

__all__ = [
    "NotNull",
    "not_null",
    "Optional",
    "optional",
]
